Index: src/x64/full-codegen-x64.cc |
diff --git a/src/x64/full-codegen-x64.cc b/src/x64/full-codegen-x64.cc |
index ed4896d8186326d12cdef5cdec78f366141b11a1..62c6073a2fe96c333f07192ac156ed8b88aaa048 100644 |
--- a/src/x64/full-codegen-x64.cc |
+++ b/src/x64/full-codegen-x64.cc |
@@ -1046,9 +1046,8 @@ void FullCodeGenerator::VisitForInStatement(ForInStatement* stmt) { |
ForIn loop_statement(this, stmt); |
increment_loop_depth(); |
// Get the object to enumerate over. Both SpiderMonkey and JSC |
// ignore null and undefined in contrast to the specification; see |
// ECMA-262 section 12.6.4. |
VisitForAccumulatorValue(stmt->enumerable()); |
__ CompareRoot(rax, Heap::kUndefinedValueRootIndex); |
__ j(equal, &exit); |
@@ -1224,6 +1223,64 @@ void FullCodeGenerator::VisitForInStatement(ForInStatement* stmt) { |
} |
+void FullCodeGenerator::VisitForOfStatement(ForOfStatement* stmt) { |
+ Comment cmnt(masm_, "[ ForOfStatement"); |
+ SetStatementPosition(stmt); |
+ |
+ Iteration loop_statement(this, stmt); |
+ increment_loop_depth(); |
+ |
+ // var iterator = iterable[@@iterator]() |
+ VisitForAccumulatorValue(stmt->assign_iterator()); |
+ |
+ // As with for-in, skip the loop if the iterator is null or undefined. |
+ __ CompareRoot(rax, Heap::kUndefinedValueRootIndex); |
+ __ j(equal, loop_statement.break_label()); |
+ __ CompareRoot(rax, Heap::kNullValueRootIndex); |
+ __ j(equal, loop_statement.break_label()); |
+ |
+ // Convert the iterator to a JS object. |
+ Label convert, done_convert; |
+ __ JumpIfSmi(rax, &convert); |
+ __ CmpObjectType(rax, FIRST_SPEC_OBJECT_TYPE, rcx); |
+ __ j(above_equal, &done_convert); |
+ __ bind(&convert); |
+ __ push(rax); |
+ __ InvokeBuiltin(Builtins::TO_OBJECT, CALL_FUNCTION); |
+ __ bind(&done_convert); |
+ |
+ // Loop entry. |
+ __ bind(loop_statement.continue_label()); |
+ |
+ // result = iterator.next() |
+ VisitForEffect(stmt->next_result()); |
+ |
+ // if (result.done) break; |
+ Label result_not_done; |
+ VisitForControl(stmt->result_done(), |
+ loop_statement.break_label(), |
+ &result_not_done, |
+ &result_not_done); |
+ __ bind(&result_not_done); |
+ |
+ // each = result.value |
+ VisitForEffect(stmt->assign_each()); |
+ |
+ // Generate code for the body of the loop. |
+ Visit(stmt->body()); |
+ |
+ // Check stack before looping. |
+ PrepareForBailoutForId(stmt->BackEdgeId(), NO_REGISTERS); |
+ EmitBackEdgeBookkeeping(stmt, loop_statement.continue_label()); |
+ __ jmp(loop_statement.continue_label()); |
+ |
+ // Exit and decrement the loop depth. |
+ PrepareForBailoutForId(stmt->ExitId(), NO_REGISTERS); |
+ __ bind(loop_statement.break_label()); |
+ decrement_loop_depth(); |
+} |
+ |
+ |
void FullCodeGenerator::EmitNewClosure(Handle<SharedFunctionInfo> info, |
bool pretenure) { |
// Use the fast case closure allocation code that allocates in new |
@@ -1046,9 +1046,8 @@ void FullCodeGenerator::VisitForInStatement(ForInStatement* stmt) { |
ForIn loop_statement(this, stmt); |
increment_loop_depth(); |
- // Get the object to enumerate over. Both SpiderMonkey and JSC |
- // ignore null and undefined in contrast to the specification; see |
- // ECMA-262 section 12.6.4. |
+ // Get the object to enumerate over. If the object is null or undefined, skip |
+ // over the loop. See ECMA-262 version 5, section 12.6.4. |
VisitForAccumulatorValue(stmt->enumerable()); |
__ CompareRoot(rax, Heap::kUndefinedValueRootIndex); |